/* APACHE Intrusion Detection Module by burak@hacettepe.edu.tr  */
/* Version 1.0							*/
/* Thanks for Sezgin Kose (sezgink@hacettepe.edu.tr) for	*/
/* helping me explore details of the Apache module API		*/

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "util_uri.h"
#include "http_core.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#define SIGNATURE_LENGTH        50
#define ID_MATCH        0
#define ID_NOMATCH      1

/* #define DEBUG		1 */

module intrusion_detection_module;

typedef struct {
                char **list;
                int nlist;
        } id_state;

int count_lines(FILE *f)
{
        char buffer[SIGNATURE_LENGTH+2];
        int i=0;
        while(1) {
                fgets(buffer,SIGNATURE_LENGTH,f);
                if (feof(f)) break;
                i++;
        }
        fseek(f,0L,SEEK_SET);
        return i;
}

int id_search(char *pattern,id_state *state)
{
        char buffer[DEFAULT_LIMIT_REQUEST_LINE+2];
        int i;
        strncpy(buffer,pattern,sizeof(buffer));
	buffer[strlen(buffer)]='\0';
        strcat(buffer,".");
#ifdef DEBUG
	ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_ALERT,NULL,
		"buffer=|%s|",buffer);
#endif
        for(i=0;i<state->nlist;i++)
                if (strncmp(buffer,state->list[i],strlen(state->list[i]))==0)
                        return ID_MATCH;
        return ID_NOMATCH;
}

static void init_id_module(server_rec *s,pool *p)
{
        FILE *f;
        int j=0;
        char buffer[SIGNATURE_LENGTH+2];

        id_state *state=ap_get_module_config(s->module_config,
                        &intrusion_detection_module);
        char *fname = ap_server_root_relative(p,"conf/config.id");

        if ((f=fopen(fname,"r"))==NULL) {
                ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_ALERT,s,
                        "Could not open ID config file");
                exit(1);
        }

        state->nlist=count_lines(f);

        state->list=(char **) ap_palloc(p,sizeof(char *)*state->nlist);
        for(j=0;j<state->nlist;j++)
                state->list[j]=
                        (char *)ap_palloc(p,sizeof(char)*SIGNATURE_LENGTH);

        for(j=0;j<state->nlist;j++) {
                fgets(buffer,SIGNATURE_LENGTH,f);
                if (buffer[strlen(buffer)-1]=='\n')
                        buffer[strlen(buffer)-1]='\0';
                strncpy(state->list[j],buffer,SIGNATURE_LENGTH);
        }
        
}

static void *make_intrusion_detection_state(pool *p, server_rec *s)
{
        id_state *state=ap_palloc(p,sizeof(id_state));
        state->list=NULL;
        state->nlist=0;
        return (void *)state;
}

static int check_intrusion(request_rec *orig)
{
        id_state *state=ap_get_module_config(orig->server->module_config,
                        &intrusion_detection_module);

        if (id_search(orig->parsed_uri.path,state)==ID_MATCH) {
                syslog(LOG_WARNING,"Attack to %s%s from %s",
                        orig->server->server_hostname,
                        orig->parsed_uri.path,
                        orig->connection->remote_ip
                );
		#ifdef DEBUG
			ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_ALERT,NULL,
				"No match with |%s|",orig->parsed_uri.path);
		#endif
		return DECLINED;
	}
	#ifdef DEBUG
			ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_ALERT,NULL,
				"Match with |%s|",orig->parsed_uri.path);
	#endif
    return OK;
}

module intrusion_detection_module=
{
    STANDARD_MODULE_STUFF,
    init_id_module,           /* initializer */
    NULL,                       /* create per-dir config */
    NULL,                       /* merge per-dir config */
    make_intrusion_detection_state,     /* server config */
    NULL,                       /* merge server config */
    NULL,                      /* command table */
    NULL,                /* handlers */
    NULL,                       /* filename translation */
    NULL,                       /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    check_intrusion,	        /* logger */
    NULL,                       /* header parser */
    NULL,                       /* child_init */
    NULL,                       /* child_exit */
    NULL                        /* post read-request */
};
